import { parse, tokenizers } from 'comment-parser';
import { format } from 'prettier';
import BSearch from 'binary-searching';
import { fromMarkdown } from 'mdast-util-from-markdown';
import parserBabel from 'prettier/plugins/babel';
import parserFlow from 'prettier/plugins/flow';
import parserTypescript from 'prettier/plugins/typescript';

const ABSTRACT = "abstract";
const ASYNC = "async";
const AUGMENTS = "augments";
const AUTHOR = "author";
const BORROWS = "borrows";
const CALLBACK = "callback";
const CATEGORY = "category";
const CLASS = "class";
const CONSTANT = "constant";
const DEFAULT = "default";
const DEFAULT_VALUE = "defaultValue";
const DEPRECATED = "deprecated";
const DESCRIPTION = "description";
const EXAMPLE = "example";
const EXTENDS = "extends";
const EXTERNAL = "external";
const FILE = "file";
const FIRES = "fires";
const FLOW = "flow";
const FUNCTION = "function";
const IGNORE = "ignore";
const LICENSE = "license";
const MEMBER = "member";
const MEMBEROF = "memberof";
const MODULE = "module";
const NAMESPACE = "namespace";
const OVERRIDE = "override";
const PARAM = "param";
const PRIVATE = "private";
const PRIVATE_REMARKS = "privateRemarks";
const PROPERTY = "property";
const PROVIDES_MODULE = "providesModule";
const REMARKS = "remarks";
const RETURNS = "returns";
const SEE = "see";
const SINCE = "since";
const TEMPLATE = "template";
const THROWS = "throws";
const TODO = "todo";
const TYPE = "type";
const TYPE_PARAM = "typeParam";
const TYPEDEF = "typedef";
const VERSION = "version";
const YIELDS = "yields";
const SPACE_TAG_DATA = {
    tag: "this_is_for_space",
    name: "",
    optional: false,
    type: "",
    description: "",
    source: [],
    problems: [],
};

const TAGS_SYNONYMS = {
    arg: PARAM,
    argument: PARAM,
    const: CONSTANT,
    constructor: CLASS,
    desc: DESCRIPTION,
    emits: FIRES,
    examples: EXAMPLE,
    exception: THROWS,
    fileoverview: FILE,
    func: FUNCTION,
    host: EXTERNAL,
    method: FUNCTION,
    overview: FILE,
    params: PARAM,
    prop: PROPERTY,
    return: RETURNS,
    var: MEMBER,
    virtual: ABSTRACT,
    yield: YIELDS,
    hidden: IGNORE,
};
const TAGS_NAMELESS = [
    BORROWS,
    CATEGORY,
    DEPRECATED,
    DESCRIPTION,
    EXAMPLE,
    EXTENDS,
    LICENSE,
    MODULE,
    NAMESPACE,
    OVERRIDE,
    PRIVATE_REMARKS,
    REMARKS,
    RETURNS,
    SINCE,
    THROWS,
    TODO,
    YIELDS,
    FILE,
];
const TAGS_TYPELESS = [
    BORROWS,
    BORROWS,
    DEPRECATED,
    DESCRIPTION,
    EXAMPLE,
    IGNORE,
    LICENSE,
    MODULE,
    NAMESPACE,
    OVERRIDE,
    PRIVATE_REMARKS,
    REMARKS,
    SINCE,
    TODO,
    FILE,
];
const TAGS_DEFAULT = [
    DEFAULT,
    DEFAULT_VALUE,
];
const TAGS_PEV_FORMATE_DESCRIPTION = [
    BORROWS,
    ...TAGS_DEFAULT,
    MEMBEROF,
    MODULE,
    SEE,
];
const TAGS_DESCRIPTION_NEEDED = [
    BORROWS,
    CATEGORY,
    DESCRIPTION,
    EXAMPLE,
    PRIVATE_REMARKS,
    REMARKS,
    SINCE,
    TODO,
];
const TAGS_VERTICALLY_ALIGN_ABLE = [
    EXTENDS,
    PARAM,
    PROPERTY,
    RETURNS,
    THROWS,
    TYPE,
    TYPEDEF,
    YIELDS,
];
const TAGS_GROUP_HEAD = [CALLBACK, TYPEDEF];
const TAGS_GROUP_CONDITION = [
    ...TAGS_GROUP_HEAD,
    TYPE,
    PROPERTY,
    PARAM,
    RETURNS,
    YIELDS,
    THROWS,
];
const TAGS_ORDER = [
    REMARKS,
    PRIVATE_REMARKS,
    PROVIDES_MODULE,
    MODULE,
    LICENSE,
    FLOW,
    ASYNC,
    PRIVATE,
    IGNORE,
    MEMBEROF,
    VERSION,
    FILE,
    AUTHOR,
    DEPRECATED,
    SINCE,
    CATEGORY,
    DESCRIPTION,
    EXAMPLE,
    ABSTRACT,
    AUGMENTS,
    CONSTANT,
    ...TAGS_DEFAULT,
    EXTERNAL,
    FIRES,
    TEMPLATE,
    TYPE_PARAM,
    FUNCTION,
    NAMESPACE,
    BORROWS,
    CLASS,
    EXTENDS,
    MEMBER,
    TYPEDEF,
    TYPE,
    PROPERTY,
    CALLBACK,
    PARAM,
    YIELDS,
    RETURNS,
    THROWS,
    "other",
    SEE,
    TODO,
];

var __awaiter$3 = (undefined && undefined.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
function convertToModernType(oldType) {
    return withoutStrings(oldType, (type) => {
        type = type.trim();
        type = type.replace(/\.</g, "<");
        type = type.replace(/\*/g, " any ");
        type = type
            .replace(/^\?\s*(\w+)$/, "$1 | null")
            .replace(/^(\w+)\s*\?$/, "$1 | null");
        let changed = true;
        while (changed) {
            changed = false;
            type = type.replace(/(^|[^$\w\xA0-\uFFFF])Array\s*<((?:[^<>=]|=>|=(?!>)|<(?:[^<>=]|=>|=(?!>))+>)+)>/g, (_, prefix, inner) => {
                changed = true;
                return `${prefix}(${inner})[]`;
            });
        }
        return type;
    });
}
function withoutStrings(type, mapFn) {
    const strings = [];
    let modifiedType = type.replace(/(["'])(?:\\(?:\r\n|[\s\S])|(?!\1)[^\\\r\n])*\1/g, (m) => {
        strings.push(m);
        return `String$${strings.length - 1}$`;
    });
    if (modifiedType.includes("`")) {
        return type;
    }
    modifiedType = mapFn(modifiedType);
    return modifiedType.replace(/String\$(\d+)\$/g, (_, index) => strings[index]);
}
function formatType(type, options) {
    return __awaiter$3(this, void 0, void 0, function* () {
        try {
            const TYPE_START = "type name = ";
            let pretty = type;
            let rest = false;
            if (pretty.startsWith("...")) {
                rest = true;
                pretty = `(${pretty.slice(3)})[]`;
            }
            pretty = yield format(`${TYPE_START}${pretty}`, Object.assign(Object.assign({}, options), { parser: "typescript", plugins: [], filepath: "file.ts" }));
            pretty = pretty.slice(TYPE_START.length);
            pretty = pretty.replace(/^\s*/g, "");
            pretty = pretty.replace(/[;\n]*$/g, "");
            pretty = pretty.trim();
            if (rest) {
                pretty = "..." + pretty.replace(/\[\s*\]$/, "");
            }
            return pretty;
        }
        catch (error) {
            return type;
        }
    });
}
function addStarsToTheBeginningOfTheLines(comment, options) {
    if (options.jsdocSingleLineComment &&
        numberOfAStringInString(comment.trim(), "\n") === 0) {
        return `* ${comment.trim()} `;
    }
    return `*${comment.replace(/(\n(?!$))/g, "\n * ")}\n `;
}
function numberOfAStringInString(string, search) {
    return (string.match(new RegExp(search, "g")) || []).length;
}
function capitalizer(str) {
    if (!str) {
        return str;
    }
    if (str.match(/^https?:\/\//i)) {
        return str;
    }
    if (str.startsWith("- ")) {
        return str.slice(0, 2) + capitalizer(str.slice(2));
    }
    return str[0].toUpperCase() + str.slice(1);
}
function detectEndOfLine(text) {
    const counter = {
        "\r": 0,
        "\r\n": 0,
        "\n": 0,
    };
    const lineEndPattern = /\r\n?|\n/g;
    let m;
    while ((m = lineEndPattern.exec(text))) {
        counter[m[0]]++;
    }
    const cr = counter["\r"];
    const crlf = counter["\r\n"];
    const lf = counter["\n"];
    const max = Math.max(cr, crlf, lf);
    if (lf === max) {
        return "lf";
    }
    else if (crlf === max) {
        return "crlf";
    }
    else {
        return "cr";
    }
}
function findTokenIndex(tokens, token) {
    return BSearch.eq(tokens, token, (a, b) => {
        if (a.loc.start.line === b.loc.start.line) {
            return a.loc.start.column - b.loc.start.column;
        }
        else {
            return a.loc.start.line - b.loc.start.line;
        }
    });
}
function formatCode(result, beginningSpace, options) {
    return __awaiter$3(this, void 0, void 0, function* () {
        const { printWidth, jsdocKeepUnParseAbleExampleIndent } = options;
        if (result
            .split("\n")
            .slice(1)
            .every((v) => !v.trim() || v.startsWith(beginningSpace))) {
            result = result.replace(new RegExp(`\n${beginningSpace.replace(/[\t]/g, "[\\t]")}`, "g"), "\n");
        }
        try {
            let formattedExample = "";
            const examplePrintWith = printWidth - 4;
            if (result.trim().startsWith("{")) {
                formattedExample = yield format(result || "", Object.assign(Object.assign({}, options), { parser: "json", printWidth: examplePrintWith }));
            }
            else {
                formattedExample = yield format(result || "", Object.assign(Object.assign({}, options), { printWidth: examplePrintWith }));
            }
            result = formattedExample.replace(/(^|\n)/g, `\n${beginningSpace}`);
        }
        catch (err) {
            result = `\n${result
                .split("\n")
                .map((l) => `${beginningSpace}${jsdocKeepUnParseAbleExampleIndent ? l : l.trim()}`)
                .join("\n")}\n`;
        }
        return result;
    });
}
const findPluginByParser = (parserName, options) => {
    var _a, _b;
    const tsPlugin = options.plugins.find((plugin) => {
        return (typeof plugin === "object" &&
            plugin.name &&
            plugin.parsers &&
            plugin.parsers.hasOwnProperty(parserName));
    });
    return !tsPlugin ||
        tsPlugin.name === "prettier-plugin-jsdoc" ||
        ((_a = tsPlugin.parsers) === null || _a === void 0 ? void 0 : _a.hasOwnProperty("jsdoc-parser"))
        ? undefined
        : (_b = tsPlugin.parsers) === null || _b === void 0 ? void 0 : _b[parserName];
};
const isDefaultTag = (tag) => TAGS_DEFAULT.includes(tag);

var __awaiter$2 = (undefined && undefined.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
const TABLE = "2@^5!~#sdE!_TABLE";
const parserSynonyms = (lang) => {
    switch (lang) {
        case "js":
        case "javascript":
        case "jsx":
            return ["babel", "babel-flow", "vue"];
        case "ts":
        case "typescript":
        case "tsx":
            return ["typescript", "babel-ts", "angular"];
        case "json":
        case "css":
            return ["css"];
        case "less":
            return ["less"];
        case "scss":
            return ["scss"];
        case "html":
            return ["html"];
        case "yaml":
            return ["yaml"];
        default:
            return ["babel"];
    }
};
function descriptionEndLine({ tag, isEndTag, }) {
    if ([DESCRIPTION, EXAMPLE, TODO].includes(tag) && !isEndTag) {
        return "\n";
    }
    return "";
}
function formatDescription(tag, text, options, formatOptions) {
    return __awaiter$2(this, void 0, void 0, function* () {
        if (!text)
            return text;
        const { printWidth } = options;
        const { tagStringLength = 0, beginningSpace } = formatOptions;
        text = text.replace(/^(\d+)[-][\s|]+/g, "$1. ");
        text = text.replace(/\n+(\s*\d+)[-][\s]+/g, "\n$1. ");
        const fencedCodeBlocks = text.matchAll(/```\S*?\n[\s\S]+?```/gm);
        const indentedCodeBlocks = text.matchAll(/^\r?\n^(?:(?:(?:[ ]{4}|\t).*(?:\r?\n|$))+)/gm);
        const allCodeBlocks = [...fencedCodeBlocks, ...indentedCodeBlocks];
        const tables = [];
        text = text.replace(/((\n|^)\|[\s\S]*?)((\n[^|])|$)/g, (code, _1, _2, _3, _, offs) => {
            for (const block of allCodeBlocks) {
                if (block.index !== undefined &&
                    block.index <= offs + 1 &&
                    offs + code.length + 1 <= block.index + block[0].length) {
                    return code;
                }
            }
            code = _3 ? code.slice(0, -1) : code;
            tables.push(code);
            return `\n\n${TABLE}\n\n${_3 ? _3.slice(1) : ""}`;
        });
        if (options.jsdocCapitalizeDescription &&
            !TAGS_PEV_FORMATE_DESCRIPTION.includes(tag)) {
            text = capitalizer(text);
        }
        text = `${tagStringLength ? `${"!".repeat(tagStringLength - 1)}?` : ""}${text.startsWith("```") ? "\n" : ""}${text}`;
        let tableIndex = 0;
        const rootAst = fromMarkdown(text);
        function stringifyASTWithoutChildren(mdAst, intention, parent) {
            return __awaiter$2(this, void 0, void 0, function* () {
                if (mdAst.type === "inlineCode") {
                    return `\`${mdAst.value}\``;
                }
                if (mdAst.type === "code") {
                    let result = mdAst.value || "";
                    let _intention = intention;
                    if (result) {
                        if (mdAst.lang) {
                            const supportParsers = parserSynonyms(mdAst.lang.toLowerCase());
                            const parser = (supportParsers === null || supportParsers === void 0 ? void 0 : supportParsers.includes(options.parser))
                                ? options.parser
                                : (supportParsers === null || supportParsers === void 0 ? void 0 : supportParsers[0]) || mdAst.lang;
                            result = yield formatCode(result, intention, Object.assign(Object.assign({}, options), { parser, jsdocKeepUnParseAbleExampleIndent: true }));
                        }
                        else if (options.jsdocPreferCodeFences || false) {
                            result = yield formatCode(result, _intention, Object.assign(Object.assign({}, options), { jsdocKeepUnParseAbleExampleIndent: true }));
                        }
                        else {
                            _intention = intention + " ".repeat(4);
                            result = yield formatCode(result, _intention, Object.assign(Object.assign({}, options), { jsdocKeepUnParseAbleExampleIndent: true }));
                        }
                    }
                    const addFence = options.jsdocPreferCodeFences || !!mdAst.lang;
                    result = addFence ? result : result.trimEnd();
                    return result
                        ? addFence
                            ? `\n\n${_intention}\`\`\`${mdAst.lang || ""}${result}\`\`\``
                            : `\n${result}`
                        : "";
                }
                if (mdAst.value === TABLE) {
                    if (parent) {
                        parent.costumeType = TABLE;
                    }
                    if (tables.length > 0) {
                        let result = (tables === null || tables === void 0 ? void 0 : tables[tableIndex]) || "";
                        tableIndex++;
                        if (result) {
                            result = (yield format(result, Object.assign(Object.assign({}, options), { parser: "markdown" }))).trim();
                        }
                        return `${result
                            ? `\n\n${intention}${result.split("\n").join(`\n${intention}`)}`
                            : mdAst.value}`;
                    }
                }
                if (mdAst.type === "break") {
                    return `\\\n`;
                }
                return (mdAst.value ||
                    mdAst.title ||
                    mdAst.alt ||
                    "");
            });
        }
        function stringyfy(mdAst, intention, parent) {
            return __awaiter$2(this, void 0, void 0, function* () {
                if (!Array.isArray(mdAst.children)) {
                    return stringifyASTWithoutChildren(mdAst, intention, parent);
                }
                return (yield Promise.all(mdAst.children
                    .map((ast, index) => __awaiter$2(this, void 0, void 0, function* () {
                    var _a;
                    switch (ast.type) {
                        case "listItem": {
                            let _listCount = `\n${intention}- `;
                            if (typeof mdAst.start === "number") {
                                const count = index + ((_a = mdAst.start) !== null && _a !== void 0 ? _a : 1);
                                _listCount = `\n${intention}${count}. `;
                            }
                            const _intention = intention + " ".repeat(_listCount.length - 1);
                            const result = (yield stringyfy(ast, _intention, mdAst)).trim();
                            return `${_listCount}${result}`;
                        }
                        case "list": {
                            let end = "";
                            if (tag !== DESCRIPTION &&
                                mdAst.type === "root" &&
                                index === mdAst.children.length - 1) {
                                end = "\n";
                            }
                            return `\n${yield stringyfy(ast, intention, mdAst)}${end}`;
                        }
                        case "paragraph": {
                            const paragraph = yield stringyfy(ast, intention, parent);
                            if (ast.costumeType === TABLE) {
                                return paragraph;
                            }
                            return `\n\n${paragraph
                                .split("\\\n")
                                .map((_paragraph) => {
                                const links = [];
                                _paragraph = _paragraph.replace(/{@(link|linkcode|linkplain)[\s](([^{}])*)}/g, (_, tag, link) => {
                                    links.push(link);
                                    return `{@${tag}${"_".repeat(link.length)}}`;
                                });
                                _paragraph = _paragraph.replace(/\s+/g, " ");
                                if (options.jsdocCapitalizeDescription &&
                                    !TAGS_PEV_FORMATE_DESCRIPTION.includes(tag))
                                    _paragraph = capitalizer(_paragraph);
                                if (options.jsdocDescriptionWithDot)
                                    _paragraph = _paragraph.replace(/([\w\p{L}])$/u, "$1.");
                                let result = breakDescriptionToLines(_paragraph, printWidth, intention);
                                result = result.replace(/{@(link|linkcode|linkplain)([_]+)}/g, (original, tag, underline) => {
                                    const link = links[0];
                                    if (link.length === underline.length) {
                                        links.shift();
                                        return `{@${tag} ${link}}`;
                                    }
                                    return original;
                                });
                                return result;
                            })
                                .join("\\\n")}`;
                        }
                        case "strong": {
                            return `**${yield stringyfy(ast, intention, mdAst)}**`;
                        }
                        case "emphasis": {
                            return `_${yield stringyfy(ast, intention, mdAst)}_`;
                        }
                        case "heading": {
                            return `\n\n${intention}${"#".repeat(ast.depth)} ${yield stringyfy(ast, intention, mdAst)}`;
                        }
                        case "link":
                        case "image": {
                            return `[${yield stringyfy(ast, intention, mdAst)}](${ast.url})`;
                        }
                        case "linkReference": {
                            return `[${yield stringyfy(ast, intention, mdAst)}][${ast.label}]`;
                        }
                        case "definition": {
                            return `\n\n[${ast.label}]: ${ast.url}`;
                        }
                        case "blockquote": {
                            const paragraph = yield stringyfy(ast, "", mdAst);
                            return `${intention}> ${paragraph
                                .trim()
                                .replace(/(\n+)/g, `$1${intention}> `)}`;
                        }
                    }
                    return stringyfy(ast, intention, mdAst);
                })))).join("");
            });
        }
        let result = yield stringyfy(rootAst, beginningSpace, null);
        result = result.replace(/^[\s\n]+/g, "");
        result = result.replace(/^([!]+\?)/g, "");
        return result;
    });
}
function breakDescriptionToLines(desContent, maxWidth, beginningSpace) {
    let str = desContent.trim();
    if (!str) {
        return str;
    }
    let result = "";
    while (str.length > maxWidth) {
        let sliceIndex = str.lastIndexOf(" ", str.startsWith("\n") ? maxWidth + 1 : maxWidth);
        if (sliceIndex <= beginningSpace.length)
            sliceIndex = str.indexOf(" ", beginningSpace.length + 1);
        if (sliceIndex === -1)
            sliceIndex = str.length;
        result += str.substring(0, sliceIndex);
        str = str.substring(sliceIndex + 1);
        if (str) {
            str = `${beginningSpace}${str}`;
            str = `\n${str}`;
        }
    }
    result += str;
    return `${beginningSpace}${result}`;
}

var __awaiter$1 = (undefined && undefined.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
const stringify = ({ name, description, type, tag }, tagIndex, finalTagsArray, options, maxTagTitleLength, maxTagTypeNameLength, maxTagNameLength) => __awaiter$1(void 0, void 0, void 0, function* () {
    let tagString = "\n";
    if (tag === SPACE_TAG_DATA.tag) {
        return tagString;
    }
    const { printWidth, jsdocSpaces, jsdocVerticalAlignment, jsdocDescriptionTag, tsdoc, useTabs, tabWidth, } = options;
    const gap = " ".repeat(jsdocSpaces);
    let tagTitleGapAdj = 0;
    let tagTypeGapAdj = 0;
    let tagNameGapAdj = 0;
    let descGapAdj = 0;
    if (jsdocVerticalAlignment && TAGS_VERTICALLY_ALIGN_ABLE.includes(tag)) {
        if (tag)
            tagTitleGapAdj += maxTagTitleLength - tag.length;
        else if (maxTagTitleLength)
            descGapAdj += maxTagTitleLength + gap.length;
        if (type)
            tagTypeGapAdj += maxTagTypeNameLength - type.length;
        else if (maxTagTypeNameLength)
            descGapAdj += maxTagTypeNameLength + gap.length;
        if (name)
            tagNameGapAdj += maxTagNameLength - name.length;
        else if (maxTagNameLength)
            descGapAdj = maxTagNameLength + gap.length;
    }
    const useTagTitle = tag !== DESCRIPTION || jsdocDescriptionTag;
    if (useTagTitle) {
        tagString += `@${tag}${" ".repeat(tagTitleGapAdj || 0)}`;
    }
    if (type) {
        const getUpdatedType = () => {
            if (!isDefaultTag(tag)) {
                return `{${type}}`;
            }
            if (type === '[]')
                return '[ ]';
            if (type === '{}')
                return '{ }';
            const isAnObject = (value) => /^{.*[A-z0-9_]+ ?:.*}$/.test(value);
            const fixObjectCommas = (objWithBrokenCommas) => objWithBrokenCommas.replace(/; ([A-z0-9_])/g, ', $1');
            if (isAnObject(type)) {
                return fixObjectCommas(type);
            }
            return type;
        };
        const updatedType = getUpdatedType();
        tagString += gap + updatedType + " ".repeat(tagTypeGapAdj);
    }
    if (name)
        tagString += `${gap}${name}${" ".repeat(tagNameGapAdj)}`;
    if (tag === EXAMPLE && !tsdoc) {
        const exampleCaption = description.match(/<caption>([\s\S]*?)<\/caption>/i);
        if (exampleCaption) {
            description = description.replace(exampleCaption[0], "");
            tagString = `${tagString} ${exampleCaption[0]}`;
        }
        const beginningSpace = useTabs ? "\t" : " ".repeat(tabWidth);
        const formattedExample = yield formatCode(description, beginningSpace, options);
        tagString += formattedExample
            .replace(new RegExp(`^\\n${beginningSpace
            .replace(/[\t]/g, "[\\t]")
            .replace(/[^S\r\n]/g, "[^S\\r\\n]")}\\n`), "")
            .trimEnd();
    }
    else if (description) {
        if (useTagTitle)
            tagString += gap + " ".repeat(descGapAdj);
        if (TAGS_PEV_FORMATE_DESCRIPTION.includes(tag) ||
            !TAGS_ORDER.includes(tag)) {
            tagString += description;
        }
        else {
            const [, firstWord] = /^\s*(\S+)/.exec(description) || ["", ""];
            const beginningSpace = tag === DESCRIPTION ||
                ([EXAMPLE, REMARKS, PRIVATE_REMARKS].includes(tag) && tsdoc)
                ? ""
                : "  ";
            if ((tag !== DESCRIPTION &&
                tagString.length + firstWord.length > printWidth) ||
                [REMARKS, PRIVATE_REMARKS].includes(tag)) {
                tagString +=
                    `\n${beginningSpace}` +
                        (yield formatDescription(tag, description, options, { beginningSpace }));
            }
            else {
                tagString += yield formatDescription(tag, description, options, {
                    tagStringLength: tagString.length - 1,
                    beginningSpace,
                });
            }
        }
    }
    tagString += descriptionEndLine({
        tag,
        isEndTag: tagIndex === finalTagsArray.length - 1,
    });
    return tagString;
});

var __awaiter = (undefined && undefined.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __rest = (undefined && undefined.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
};
const { name: nameTokenizer, tag: tagTokenizer, type: typeTokenizer, description: descriptionTokenizer, } = tokenizers;
const getParser = (originalParse, parserName) => function jsdocParser(text, parsersOrOptions, maybeOptions) {
    var _a, _b;
    return __awaiter(this, void 0, void 0, function* () {
        let options = (maybeOptions !== null && maybeOptions !== void 0 ? maybeOptions : parsersOrOptions);
        const prettierParse = ((_a = findPluginByParser(parserName, options)) === null || _a === void 0 ? void 0 : _a.parse) || originalParse;
        const ast = prettierParse(text, options);
        if (options.jsdocParser === false) {
            return ast;
        }
        options = Object.assign(Object.assign({}, options), { printWidth: (_b = options.jsdocPrintWidth) !== null && _b !== void 0 ? _b : options.printWidth });
        const eol = options.endOfLine === "auto" ? detectEndOfLine(text) : options.endOfLine;
        options = Object.assign(Object.assign({}, options), { endOfLine: "lf" });
        yield Promise.all(ast.comments.map((comment) => __awaiter(this, void 0, void 0, function* () {
            if (!isBlockComment(comment))
                return;
            const tokenIndex = findTokenIndex(ast.tokens, comment);
            const paramsOrder = getParamsOrders(ast, tokenIndex);
            comment.value = comment.value.replace(/^([*]+)/g, "*");
            const commentString = `/*${comment.value.replace(/\r\n?/g, "\n")}*/`;
            if (!/^\/\*\*[\s\S]+?\*\/$/.test(commentString))
                return;
            const parsed = parse(commentString, {
                spacing: "preserve",
                tokenizers: [
                    tagTokenizer(),
                    (spec) => {
                        if ([DEFAULT, DEFAULT_VALUE].includes(spec.tag)) {
                            return spec;
                        }
                        return typeTokenizer("preserve")(spec);
                    },
                    nameTokenizer(),
                    descriptionTokenizer("preserve"),
                ],
            })[0];
            comment.value = "";
            if (!parsed) {
                return;
            }
            normalizeTags(parsed);
            convertCommentDescToDescTag(parsed);
            const commentContentPrintWidth = getIndentationWidth(comment, text, options);
            let maxTagTitleLength = 0;
            let maxTagTypeLength = 0;
            let maxTagNameLength = 0;
            let tags = parsed.tags
                .map((_a) => {
                var { type, optional } = _a, rest = __rest(_a, ["type", "optional"]);
                if (type) {
                    type = type.replace(/[=]$/, () => {
                        optional = true;
                        return "";
                    });
                    type = convertToModernType(type);
                }
                return Object.assign(Object.assign({}, rest), { type,
                    optional });
            });
            tags = sortTags(tags, paramsOrder, options);
            if (options.jsdocSeparateReturnsFromParam) {
                tags = tags.flatMap((tag, index) => {
                    var _a;
                    if (tag.tag === RETURNS && ((_a = tags[index - 1]) === null || _a === void 0 ? void 0 : _a.tag) === PARAM) {
                        return [SPACE_TAG_DATA, tag];
                    }
                    return [tag];
                });
            }
            if (options.jsdocAddDefaultToDescription) {
                tags = tags.map(addDefaultValueToDescription);
            }
            tags = yield Promise.all(tags
                .map(assignOptionalAndDefaultToName)
                .map((_c) => __awaiter(this, void 0, void 0, function* () {
                var { type } = _c, rest = __rest(_c, ["type"]);
                if (type) {
                    type = yield formatType(type, Object.assign(Object.assign({}, options), { printWidth: commentContentPrintWidth }));
                }
                return Object.assign(Object.assign({}, rest), { type });
            }))).then((formattedTags) => formattedTags.map((_a) => {
                var { type, name, description, tag } = _a, rest = __rest(_a, ["type", "name", "description", "tag"]);
                const isVerticallyAlignAbleTags = TAGS_VERTICALLY_ALIGN_ABLE.includes(tag);
                if (isVerticallyAlignAbleTags) {
                    maxTagTitleLength = Math.max(maxTagTitleLength, tag.length);
                    maxTagTypeLength = Math.max(maxTagTypeLength, type.length);
                    maxTagNameLength = Math.max(maxTagNameLength, name.length);
                }
                return Object.assign({ type,
                    name, description: description.trim(), tag }, rest);
            }));
            if (options.jsdocSeparateTagGroups) {
                tags = tags.flatMap((tag, index) => {
                    const prevTag = tags[index - 1];
                    if (prevTag &&
                        prevTag.tag !== DESCRIPTION &&
                        prevTag.tag !== EXAMPLE &&
                        prevTag.tag !== SPACE_TAG_DATA.tag &&
                        tag.tag !== SPACE_TAG_DATA.tag &&
                        prevTag.tag !== tag.tag) {
                        return [SPACE_TAG_DATA, tag];
                    }
                    return [tag];
                });
            }
            const filteredTags = tags.filter(({ description, tag }) => {
                if (!description && TAGS_DESCRIPTION_NEEDED.includes(tag)) {
                    return false;
                }
                return true;
            });
            for (const [tagIndex, tagData] of filteredTags.entries()) {
                const formattedTag = yield stringify(tagData, tagIndex, filteredTags, Object.assign(Object.assign({}, options), { printWidth: commentContentPrintWidth }), maxTagTitleLength, maxTagTypeLength, maxTagNameLength);
                comment.value += formattedTag;
            }
            comment.value = comment.value.trimEnd();
            if (comment.value) {
                comment.value = addStarsToTheBeginningOfTheLines(comment.value, options);
            }
            if (eol === "cr") {
                comment.value = comment.value.replace(/\n/g, "\r");
            }
            else if (eol === "crlf") {
                comment.value = comment.value.replace(/\n/g, "\r\n");
            }
        })));
        ast.comments = ast.comments.filter((comment) => !(isBlockComment(comment) && !comment.value));
        return ast;
    });
};
function sortTags(tags, paramsOrder, options) {
    let canGroupNextTags = false;
    let shouldSortAgain = false;
    tags = tags
        .reduce((tagGroups, cur) => {
        if (tagGroups.length === 0 ||
            (TAGS_GROUP_HEAD.includes(cur.tag) && canGroupNextTags)) {
            canGroupNextTags = false;
            tagGroups.push([]);
        }
        if (TAGS_GROUP_CONDITION.includes(cur.tag)) {
            canGroupNextTags = true;
        }
        tagGroups[tagGroups.length - 1].push(cur);
        return tagGroups;
    }, [])
        .flatMap((tagGroup, index, array) => {
        var _a;
        tagGroup.sort((a, b) => {
            if (paramsOrder &&
                paramsOrder.length > 1 &&
                a.tag === PARAM &&
                b.tag === PARAM) {
                const aIndex = paramsOrder.indexOf(a.name);
                const bIndex = paramsOrder.indexOf(b.name);
                if (aIndex > -1 && bIndex > -1) {
                    return aIndex - bIndex;
                }
                return 0;
            }
            return (getTagOrderWeight(a.tag, options) - getTagOrderWeight(b.tag, options));
        });
        if (array.length - 1 !== index) {
            tagGroup.push(SPACE_TAG_DATA);
        }
        if (index > 0 &&
            ((_a = tagGroup[0]) === null || _a === void 0 ? void 0 : _a.tag) &&
            !TAGS_GROUP_HEAD.includes(tagGroup[0].tag)) {
            shouldSortAgain = true;
        }
        return tagGroup;
    });
    return shouldSortAgain ? sortTags(tags, paramsOrder, options) : tags;
}
function getTagOrderWeight(tag, options) {
    if (tag === DESCRIPTION && !options.jsdocDescriptionTag) {
        return -1;
    }
    const index = TAGS_ORDER.indexOf(tag);
    return index === -1 ? TAGS_ORDER.indexOf("other") : index;
}
function isBlockComment(comment) {
    return comment.type === "CommentBlock" || comment.type === "Block";
}
function getIndentationWidth(comment, text, options) {
    const line = text.split(/\r\n?|\n/g)[comment.loc.start.line - 1];
    let spaces = 0;
    let tabs = 0;
    for (let i = comment.loc.start.column - 1; i >= 0; i--) {
        const c = line[i];
        if (c === " ") {
            spaces++;
        }
        else if (c === "\t") {
            tabs++;
        }
        else {
            break;
        }
    }
    return options.printWidth - (spaces + tabs * options.tabWidth) - " * ".length;
}
const TAGS_ORDER_LOWER = TAGS_ORDER.map((tagOrder) => tagOrder.toLowerCase());
function normalizeTags(parsed) {
    parsed.tags = parsed.tags.map((_a) => {
        var { tag, type, name, description, default: _default } = _a, rest = __rest(_a, ["tag", "type", "name", "description", "default"]);
        tag = tag || "";
        type = type || "";
        name = name || "";
        description = description || "";
        _default = _default === null || _default === void 0 ? void 0 : _default.trim();
        const tagSticksToType = tag.indexOf("{");
        if (tagSticksToType !== -1 && tag[tag.length - 1] === "}") {
            type = tag.slice(tagSticksToType + 1, -1) + " " + type;
            tag = tag.slice(0, tagSticksToType);
        }
        tag = tag.trim();
        const lower = tag.toLowerCase();
        const tagIndex = TAGS_ORDER_LOWER.indexOf(lower);
        if (tagIndex >= 0) {
            tag = TAGS_ORDER[tagIndex];
        }
        else if (lower in TAGS_SYNONYMS) {
            tag = TAGS_SYNONYMS[lower];
        }
        type = type.trim();
        name = name.trim();
        if (name && TAGS_NAMELESS.includes(tag)) {
            description = `${name} ${description}`;
            name = "";
        }
        if (type && TAGS_TYPELESS.includes(tag)) {
            description = `{${type}} ${description}`;
            type = "";
        }
        description = description.trim();
        return Object.assign({ tag,
            type,
            name,
            description, default: _default }, rest);
    });
}
function convertCommentDescToDescTag(parsed) {
    let description = parsed.description || "";
    parsed.description = "";
    parsed.tags = parsed.tags.filter(({ description: _description, tag }) => {
        if (tag.toLowerCase() === DESCRIPTION) {
            if (_description.trim()) {
                description += "\n\n" + _description;
            }
            return false;
        }
        else {
            return true;
        }
    });
    if (description) {
        parsed.tags.unshift({
            tag: DESCRIPTION,
            description,
            name: undefined,
            type: undefined,
            source: [],
            optional: false,
            problems: [],
        });
    }
}
function getParamsOrders(ast, tokenIndex) {
    var _a;
    let params;
    try {
        const nextTokenType = (_a = ast.tokens[tokenIndex + 1]) === null || _a === void 0 ? void 0 : _a.type;
        if (typeof nextTokenType !== "object") {
            return undefined;
        }
        if (nextTokenType.label === "function") {
            let openedParenthesesCount = 1;
            const tokensAfterComment = ast.tokens.slice(tokenIndex + 4);
            const endIndex = tokensAfterComment.findIndex(({ type }) => {
                if (typeof type === "string") {
                    return false;
                }
                else if (type.label === "(") {
                    openedParenthesesCount++;
                }
                else if (type.label === ")") {
                    openedParenthesesCount--;
                }
                return openedParenthesesCount === 0;
            });
            params = tokensAfterComment.slice(0, endIndex + 1);
        }
        if (nextTokenType.label === "const") {
            let openedParenthesesCount = 1;
            let tokensAfterComment = ast.tokens.slice(tokenIndex + 1);
            const firstParenthesesIndex = tokensAfterComment.findIndex(({ type }) => typeof type === "object" && type.label === "(");
            tokensAfterComment = tokensAfterComment.slice(firstParenthesesIndex + 1);
            const endIndex = tokensAfterComment.findIndex(({ type }) => {
                if (typeof type === "string") {
                    return false;
                }
                else if (type.label === "(") {
                    openedParenthesesCount++;
                }
                else if (type.label === ")") {
                    openedParenthesesCount--;
                }
                return openedParenthesesCount === 0;
            });
            const arrowItem = tokensAfterComment[endIndex + 1];
            if (typeof (arrowItem === null || arrowItem === void 0 ? void 0 : arrowItem.type) === "object" &&
                arrowItem.type.label === "=>") {
                params = tokensAfterComment.slice(0, endIndex + 1);
            }
        }
        return params === null || params === void 0 ? void 0 : params.filter(({ type }) => typeof type === "object" && type.label === "name").map(({ value }) => value);
    }
    catch (_b) {
    }
    return;
}
function addDefaultValueToDescription(tag) {
    if (tag.optional && tag.default) {
        let { description } = tag;
        description = description.replace(/[ \t]*Default is `.*`\.?$/, "");
        if (description && !/[.\n]$/.test(description)) {
            description += ".";
        }
        description += ` Default is \`${tag.default}\``;
        return Object.assign(Object.assign({}, tag), { description: description.trim() });
    }
    else {
        return tag;
    }
}
function assignOptionalAndDefaultToName(_a) {
    var _b;
    var { name, optional, default: default_, tag, type, source, description } = _a, rest = __rest(_a, ["name", "optional", "default", "tag", "type", "source", "description"]);
    if (isDefaultTag(tag)) {
        const usefulSourceLine = ((_b = source.find((x) => x.source.includes(`@${tag}`))) === null || _b === void 0 ? void 0 : _b.source) || "";
        const tagMatch = usefulSourceLine.match(/@default(Value)? (\[.*]|{.*}|\(.*\)|'.*'|".*"|`.*`|[A-z0-9_]+)( (.+))?/);
        const tagValue = (tagMatch === null || tagMatch === void 0 ? void 0 : tagMatch[2]) || "";
        const tagDescription = (tagMatch === null || tagMatch === void 0 ? void 0 : tagMatch[4]) || "";
        if (tagMatch) {
            type = tagValue;
            name = "";
            description = tagDescription;
        }
    }
    else if (optional) {
        if (name) {
            if (default_) {
                name = `[${name}=${default_}]`;
            }
            else {
                name = `[${name}]`;
            }
        }
        else {
            type = `${type} | undefined`;
        }
    }
    return Object.assign(Object.assign({}, rest), { tag,
        name,
        description,
        optional,
        type,
        source, default: default_ });
}

const options = {
    jsdocParser: {
        name: "jsdocParser",
        type: "boolean",
        category: "jsdoc",
        default: true,
        description: "Enable/Disable jsdoc parser",
    },
    jsdocSpaces: {
        name: "jsdocSpaces",
        type: "int",
        category: "jsdoc",
        default: 1,
        description: "How many spaces will be used to separate tag elements.",
    },
    jsdocDescriptionWithDot: {
        name: "jsdocDescriptionWithDot",
        type: "boolean",
        category: "jsdoc",
        default: false,
        description: "Should dot be inserted at the end of description",
    },
    jsdocDescriptionTag: {
        name: "jsdocDescriptionTag",
        type: "boolean",
        category: "jsdoc",
        default: false,
        description: "Should description tag be used",
    },
    jsdocVerticalAlignment: {
        name: "jsdocVerticalAlignment",
        type: "boolean",
        category: "jsdoc",
        default: false,
        description: "Should tags, types, names and description be aligned",
    },
    jsdocKeepUnParseAbleExampleIndent: {
        name: "jsdocKeepUnParseAbleExampleIndent",
        type: "boolean",
        category: "jsdoc",
        default: false,
        description: "Should unParseAble example (pseudo code or no js code) keep its indentation",
    },
    jsdocSingleLineComment: {
        name: "jsdocSingleLineComment",
        type: "boolean",
        category: "jsdoc",
        default: true,
        description: "Should compact single line comment",
    },
    jsdocSeparateReturnsFromParam: {
        name: "jsdocSeparateReturnsFromParam",
        type: "boolean",
        category: "jsdoc",
        default: false,
        description: "Add an space between last @param and @returns",
    },
    jsdocSeparateTagGroups: {
        name: "jsdocSeparateTagGroups",
        type: "boolean",
        category: "jsdoc",
        default: false,
        description: "Add an space between tag groups",
    },
    jsdocCapitalizeDescription: {
        name: "jsdocCapitalizeDescription",
        type: "boolean",
        category: "jsdoc",
        default: true,
        description: "Should capitalize first letter of description",
    },
    tsdoc: {
        name: "tsdoc",
        type: "boolean",
        category: "jsdoc",
        default: false,
        description: "Should format as tsdoc",
    },
    jsdocPrintWidth: {
        name: "jsdocPrintWidth",
        type: "int",
        category: "jsdoc",
        default: undefined,
        description: "If You don't set value to jsdocPrintWidth, the printWidth will be use as jsdocPrintWidth.",
    },
    jsdocAddDefaultToDescription: {
        name: "jsdocAddDefaultToDescription",
        type: "boolean",
        category: "jsdoc",
        default: true,
        description: "Add Default value of a param to end description",
    },
    jsdocPreferCodeFences: {
        name: "jsdocPreferCodeFences",
        type: "boolean",
        category: "jsdoc",
        default: false,
        description: `Prefer to render code blocks using "fences" (triple backticks). If not set, blocks without a language tag will be rendered with a four space indentation.`,
    },
    jsdocLineWrappingStyle: {
        name: "jsdocLineWrappingStyle",
        type: "choice",
        choices: [
            { since: "0.3.39", value: "greedy", description: `Lines wrap as soon as they reach the print width` },
        ],
        category: "jsdoc",
        default: "greedy",
        description: `Strategy for wrapping lines for the given print width. More options may be added in the future.`,
    },
};
const defaultOptions = {
    jsdocParser: options.jsdocParser.default,
    jsdocSpaces: options.jsdocSpaces.default,
    jsdocPrintWidth: options.jsdocPrintWidth.default,
    jsdocDescriptionWithDot: options.jsdocDescriptionWithDot.default,
    jsdocDescriptionTag: options.jsdocDescriptionTag.default,
    jsdocVerticalAlignment: options.jsdocVerticalAlignment.default,
    jsdocKeepUnParseAbleExampleIndent: options.jsdocKeepUnParseAbleExampleIndent
        .default,
    jsdocSingleLineComment: options.jsdocSingleLineComment.default,
    jsdocSeparateReturnsFromParam: options.jsdocSeparateReturnsFromParam
        .default,
    jsdocSeparateTagGroups: options.jsdocSeparateTagGroups.default,
    jsdocCapitalizeDescription: options.jsdocCapitalizeDescription
        .default,
    jsdocAddDefaultToDescription: options.jsdocAddDefaultToDescription
        .default,
    jsdocPreferCodeFences: options.jsdocPreferCodeFences.default,
    tsdoc: options.tsdoc.default,
    jsdocLineWrappingStyle: options.jsdocLineWrappingStyle.default,
};
const parsers = {
    get babel() {
        const parser = parserBabel.parsers.babel;
        return mergeParsers(parser, "babel");
    },
    get "babel-flow"() {
        const parser = parserBabel.parsers["babel-flow"];
        return mergeParsers(parser, "babel-flow");
    },
    get "babel-ts"() {
        const parser = parserBabel.parsers["babel-ts"];
        return mergeParsers(parser, "babel-ts");
    },
    get flow() {
        const parser = parserFlow.parsers.flow;
        return mergeParsers(parser, "flow");
    },
    get typescript() {
        const parser = parserTypescript.parsers.typescript;
        return mergeParsers(parser, "typescript");
    },
    get "jsdoc-parser"() {
        const parser = parserBabel.parsers["babel-ts"];
        return mergeParsers(parser, "babel-ts");
    },
};
function mergeParsers(originalParser, parserName) {
    const jsDocParse = getParser(originalParser.parse, parserName);
    const jsDocPreprocess = (text, options) => {
        const tsPluginParser = findPluginByParser(parserName, options);
        if (!tsPluginParser) {
            return originalParser.preprocess
                ? originalParser.preprocess(text, options)
                : text;
        }
        const preprocess = tsPluginParser.preprocess || originalParser.preprocess;
        Object.assign(parser, Object.assign(Object.assign(Object.assign({}, parser), tsPluginParser), { preprocess: jsDocPreprocess, parse: jsDocParse }));
        return preprocess ? preprocess(text, options) : text;
    };
    const parser = Object.assign(Object.assign({}, originalParser), { preprocess: jsDocPreprocess, parse: jsDocParse });
    return parser;
}

export { defaultOptions, options, parsers };
